# -*- coding: utf-8 -*-
"""Find best fitting hrf parameters."""

# Part of hrf_opt library
# Copyright (C) 2018  Marian Schneider
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from sklearn.model_selection import KFold
from pyprf_feature.analysis.utils_hrf import spm_hrf_compat
from hrf_opt.find_hrf_utils_cy_one import (cy_lst_sq_one, cy_lst_sq_xval_one)
from hrf_opt.find_hrf_utils_cy_two import (cy_lst_sq_two, cy_lst_sq_xval_two)
from hrf_opt.hrf_opt_utils import (cnvl_nrl_hrf, cnvl_nrl_hrf_vec,
                                   cvrt_hrf_prm_fn)


def find_opt_hrf(idxPrc, aryFunc, aryNrlRsp, aryHrfBse, aryPrm, varTr,
                 varNumVol, varTmpOvsmpl, lgcXval, varNumXval, queOut,
                 lgcPrint=True):
    """
    Find best hrf parameters for voxel time course, using the CPU.

    Parameters
    ----------
    idxPrc : int
        Process ID of the process calling this function (for CPU
        multi-threading). In GPU version, this parameter is 0 (just one thread
        on CPU).
    aryFunc : np.array
        2D array with functional MRI data, with shape aryFunc[voxel, time].
    aryNrlRsp : np.array
        Array with best fitting neural response.
    aryHrfBse : np.array
        2D array with all basis hrf functions.
    aryPrm : np.array
        2D array with all hrf parameter combinations.
    varTr : float, positive
        Time to repeat (TR) of the (fMRI) experiment.
    varNumVol : float, positive
        Number of volumes of the (fMRI) data.
    varTmpOvsmpl : float, positive
        Factor by which the time courses has been temporally upsampled.
    lgcXval: boolean
        Logical to determine whether we cross-validate.
    varNumXval: int
        Number of folds for k-fold cross-validation.
    queOut : multiprocessing.queues.Queue
        Queue to put the results on.
    lgcPrint : boolean
        Whether print statements should be executed.

    Returns
    -------
    lstOut : list
        List containing the following objects:
        idxPrc : int
            Process ID of the process calling this function (for CPU
            multi-threading). In GPU version, this parameter is 0.
        vecBstXpos : np.array
            1D array with best fitting x-position for each voxel, with shape
            vecBstXpos[voxel].
        vecBstYpos : np.array
            1D array with best fitting y-position for each voxel, with shape
            vecBstYpos[voxel].
        vecBstSd : np.array
            1D array with best fitting pRF size for each voxel, with shape
            vecBstSd[voxel].
        vecBstR2 : np.array
            1D array with R2 value of 'winning' pRF model for each voxel, with
            shape vecBstR2[voxel].
        aryBstBts : np.array
            2D array with beta parameter estimates of 'winning' pRF model for
            each voxel, with shape aryBstBts[voxel, feautures].

    Notes
    -----
    The list with results is not returned directly, but placed on a
    multiprocessing queue. This version performs the model finding on the CPU,
    using cython.

    """

    # Derive number of features
    varNumFtr = aryNrlRsp.shape[1]

    # Derive number of voxels to be fitted in this chunk:
    varNumVoxChnk = aryFunc.shape[-1]

    # Derive number of models:
    varNumMdls = aryHrfBse.shape[0]

    # Make array to collect residuals for all voxels and models
    aryClcRes = np.zeros((varNumVoxChnk, varNumMdls)).astype(np.float32)

    # Vectors for hrf finding results.
    # Make sure they have the same precision as aryPrm, since this
    # is important for later comparison
    vecBstPrm = np.zeros((varNumVoxChnk, aryPrm.shape[-1]), dtype=aryPrm.dtype)

    # Vector for best R-square value. For each model fit, the R-square value is
    # compared to this, and updated if it is lower than the best-fitting
    # solution so far. We initialise with an arbitrary, high value
    vecBstRes = np.add(np.zeros(varNumVoxChnk), np.inf).astype(np.float32)

    # array for best beta values. If we update the residual value above because
    # it is lower, we also update the beta values of these voxels
    aryBstBts = np.zeros((varNumVoxChnk, varNumFtr)).astype(np.float32)

    # In case we cross-validate we also save and replace the best
    # residual values for every fold (not only mean across folds):
    if lgcXval:
        aryBstResFlds = np.zeros((varNumVoxChnk, varNumXval), dtype=np.float32)

    # Set type to float 32:
    aryFunc = aryFunc.astype(np.float32)

    # if lgc for Xval is true we already prepare indices for xvalidation
    if lgcXval:
        # obtain iterator for cross-validation
        itXval = KFold(n_splits=varNumXval)
        vecSplts = np.arange(aryFunc.shape[0], dtype=np.int32)

        # prepare lists that will hold indices for xvalidation
        lstIdxTrn = []
        lstIdxtst = []
        # Loop over the cross-validations to put indcies in array
        for idxTrn, idxTst in itXval.split(vecSplts):
            lstIdxTrn.append(idxTrn)
            lstIdxtst.append(idxTst)
        # trun lists into array
        aryIdxTrn = np.stack(lstIdxTrn, axis=-1).astype(np.int32)
        aryIdxTst = np.stack(lstIdxtst, axis=-1).astype(np.int32)

    # Prepare status indicator if this is the first of the parallel processes:
    if idxPrc == 0:

        # We create a status indicator for the time consuming pRF model finding
        # algorithm. Number of steps of the status indicator:
        varStsStpSze = 20

        # Vector with pRF values at which to give status feedback:
        vecStatPrf = np.linspace(0,
                                 varNumVoxChnk,
                                 num=(varStsStpSze+1),
                                 endpoint=True)
        vecStatPrf = np.ceil(vecStatPrf)
        vecStatPrf = vecStatPrf.astype(int)

        # Vector with corresponding percentage values at which to give status
        # feedback:
        vecStatPrc = np.linspace(0,
                                 100,
                                 num=(varStsStpSze+1),
                                 endpoint=True)
        vecStatPrc = np.ceil(vecStatPrc)
        vecStatPrc = vecStatPrc.astype(int)

        # Counter for status indicator:
        varCntSts01 = 0
        varCntSts02 = 0

    # Loop through voxels:
    for idxVxl in range(0, varNumVoxChnk):

        # Status indicator (only used in the first of the parallel
        # processes):
        if idxPrc == 0:

            # Status indicator:
            if varCntSts02 == vecStatPrf[varCntSts01]:

                # Prepare status message:
                strStsMsg = ('------------Progress: ' +
                             str(vecStatPrc[varCntSts01]) +
                             ' % --- ' +
                             str(vecStatPrf[varCntSts01]) +
                             ' voxels out of ' +
                             str(varNumVoxChnk))
                if lgcPrint:
                    print(strStsMsg)

                # Only increment counter if the last value has not been
                # reached yet:
                if varCntSts01 < varStsStpSze:
                    varCntSts01 = varCntSts01 + int(1)

        # For this particular voxel get best-fitting neural response
        aryVxlNrlRsp = np.atleast_2d(aryNrlRsp[idxVxl, ...]).astype(np.float32)
        # Get frame times, i.e. start point of every volume in seconds
        vecFrms = np.arange(0, varTr * varNumVol, varTr)
        # Get supersampled frames times, i.e. start point of every volume in
        # upsampled res, since convolution takes place in temp. upsampled space
        vecFrmTms = np.arange(0, varTr * varNumVol, varTr / varTmpOvsmpl)

        # Convolving the best-fitting neural response with all hrf base fn
        aryMdlRsp = cnvl_nrl_hrf(aryVxlNrlRsp, aryHrfBse, vecFrms, vecFrmTms,
                                 varTr, varNumVol)
#        aryMdlRsp = cnvl_nrl_hrf_vec(aryVxlNrlRsp, aryHrfBse, vecFrms,
#                                     vecFrmTms, varTr, varNumVol)

        # Check whether we need to crossvalidate
        if lgcXval:
            # We do crossvalidate. In this case, we loop through
            # the different folds of the crossvalidation and
            # calculate the cross-validation error for the current
            # model for all voxel time courses.

            # A cython function is used to calculate the residuals and
            # beta parameter estimates of the current model:
            if varNumFtr == 1:
                # For time course with one predictors
                aryResXval = cy_lst_sq_xval_one(np.squeeze(aryMdlRsp).T,
                                                aryFunc[:, idxVxl],
                                                aryIdxTrn,
                                                aryIdxTst)

            elif varNumFtr == 2:
                # For time course with two predictors
                aryResXval = cy_lst_sq_xval_two(
                    np.transpose(aryMdlRsp, (2, 1, 0)), aryFunc[:, idxVxl],
                    aryIdxTrn, aryIdxTst)

            else:
                if lgcPrint:
                    print('Cython currently not implemented for ' +
                          'more than two predictors.')

            # calculate the average cross validation error across
            # all folds
            vecTmpRes = np.mean(aryResXval, axis=1)

        else:
            # We do not crossvalidate. In this case, we calculate
            # the ratio of the explained variance (R squared)
            # for the current model for all voxel time courses.

            # A cython function is used to calculate the residuals and
            # beta parameter estimates of the current model:
            if varNumFtr == 1:
                # For time course with one predictor
                aryTmpBts, vecTmpRes = cy_lst_sq_one(
                    np.squeeze(aryMdlRsp).T, aryFunc[:, idxVxl])

            elif varNumFtr == 2:
                # For time course with two predictors
                aryTmpBts, vecTmpRes = cy_lst_sq_two(
                    np.transpose(aryMdlRsp, (2, 1, 0)), aryFunc[:, idxVxl])
            else:
                if lgcPrint:
                    print('Cython currently not implemented for ' +
                          'more than two two predictors.')

        # Collect residuals for this voxel for all models
        aryClcRes[idxVxl, :] = vecTmpRes

        # Get index for the winner model
        idxWnrMdl = np.argmin(vecTmpRes)

        # Replace best hrf parameters:
        vecBstPrm[idxVxl, :] = aryPrm[idxWnrMdl, :]

        # Replace best mean residual values:
        vecBstRes[idxVxl] = vecTmpRes[idxWnrMdl]

        if not lgcXval:
            # Replace best beta values:
            aryBstBts[idxVxl, :] = aryTmpBts[:, idxWnrMdl].T

        # In case we cross-validate we also save and replace the best
        # residual values for every fold (not only mean across folds):
        if lgcXval:
            aryBstResFlds[idxVxl, :] = aryResXval[idxWnrMdl, :]

        # Status indicator (only used in the first of the parallel
        # processes):
        if idxPrc == 0:

            # Increment status indicator counter:
            varCntSts02 = varCntSts02 + 1

    # After finding the best fitting model for each voxel, we still have to
    # calculate the average correlation coefficient between predicted and
    # measured time course (xval=True) or the coefficient of determination
    # (xval=False) for each voxel.

    if lgcXval:

        # Since we did not do this during finding the best model, we still need
        # to calculate deviation from a mean model for every voxel and fold
        # arySsTotXval as well as calculate the best betas for the full model

        # Calculate deviation from a mean model for every voxel and fold
        arySsTotXval = np.zeros((aryBstResFlds.shape),
                                dtype=aryBstResFlds.dtype)

        # Loop over all best-fitting model parameter combinations found
        for indPrm, vecPrm in enumerate(vecBstPrm):
            # Get voxel time course
            aryVxlTc = aryFunc[:, indPrm]
            # Get nerual model time courses
            aryNrlTmp = aryNrlRsp[indPrm, :]
            # Get hrf basis function based on best hrf params for this voxel
            aryHrfTmp = cvrt_hrf_prm_fn(np.atleast_2d(vecPrm), spm_hrf_compat,
                                        varTr, varTmpOvsmpl)
            # Create model time course
            aryMdlTc = cnvl_nrl_hrf_vec(aryNrlTmp, aryHrfTmp, vecFrms,
                                        vecFrmTms, varTr, varNumVol)

            # Calculate beta parameter estimates for entire model
            aryBstBts[indPrm, :] = np.linalg.lstsq(aryMdlTc.T,
                                                   aryVxlTc,
                                                   rcond=-1)[0]

            # loop over cross-validation folds
            for idxXval in range(varNumXval):
                # Get functional data for tst:
                aryFuncTst = aryVxlTc[aryIdxTst[:, idxXval]]
                # Deviation from the mean for each datapoint:
                aryFuncDev = np.subtract(aryFuncTst,
                                         np.mean(aryFuncTst,
                                                 axis=0))
                # Sum of squares:
                vecSsTot = np.sum(np.power(aryFuncDev, 2.0), axis=0)
                arySsTotXval[indPrm, idxXval] = vecSsTot

        # Calculate coefficient of determination by comparing:
        # aryBstResFlds vs. arySsTotXval

        # get logical to check that arySsTotXval is greater than zero in all
        # voxels and folds
        lgcExclZeros = np.all(np.greater(arySsTotXval,  np.array([0.0])),
                              axis=1)
        if lgcPrint:
            print('------------Nr of voxels: ' + str(len(lgcExclZeros)))
            print('------------Nr of voxels avove 0: ' +
                  str(np.sum(lgcExclZeros)))

        # Calculate R2 for every crossvalidation fold seperately
        aryBstR2fld = np.subtract(
            1.0, np.divide(aryBstResFlds,
                           arySsTotXval))

        # Calculate mean R2 across folds here
        vecBstR2 = np.subtract(
            1.0, np.mean(np.divide(aryBstResFlds,
                                   arySsTotXval),
                         axis=1))

        # Output list:
        lstOut = [idxPrc,
                  vecBstPrm,
                  vecBstR2,
                  aryBstBts,
                  aryClcRes,
                  aryBstR2fld]

        queOut.put(lstOut)

    else:
        # To calculate the coefficient of determination, we start with the
        # total sum of squares (i.e. the deviation of the data from the mean).
        # The mean of each time course:
        vecFuncMean = np.mean(aryFunc, axis=0)
        # Deviation from the mean for each datapoint:
        aryFuncDev = np.subtract(aryFunc, vecFuncMean[None, :])
        # Sum of squares:
        vecSsTot = np.sum(np.power(aryFuncDev,
                                   2.0),
                          axis=0)
        # Coefficient of determination:
        vecBstR2 = np.subtract(1.0,
                               np.divide(vecBstRes,
                                         vecSsTot))

        # Output list:
        lstOut = [idxPrc,
                  vecBstPrm,
                  vecBstR2,
                  aryBstBts,
                  aryClcRes]

        queOut.put(lstOut)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 23 09:05:57 2018

@author: Marian
"""

import itertools
import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import fftconvolve


def crt_hrf_prm(varPkDelMin, varPkDelMax, varUndDelMin, varUndDelMax,
                varDelStp, varPkDspMin, varPkDspMax, varUndDspMin,
                varUndDspMax, varDspStp, varPkUndRatMin, varPkUndRatMax,
                varPkUndRatStp):

    """Create combinations of hrf parameters.

    Parameters
    ----------
    varPkDelMin : float, positive
        Minimum delay for the peak of the hrf function.
    varPkDelMax : float, positive
        Maximum delay for the peak of the hrf function.
    varUndDelMin : float, positive
        Minimum delay for the undershoot of the hrf function.
    varUndDelMax : float, positive
        Maximum delay for the undershoot of the hrf function.
    varDelStp : float, positive
        Step size for peak and undershoot delay of the hrf function.
    varPkDspMin : float, positive
        Minimum dispersion for the peak of the hrf function.
    varPkDspMax : float, positive
        Maximum dispersion for the peak of the hrf function.
    varUndDspMin : float, positive
        Minimum dispersion for the undershoot of the hrf function.
    varUndDspMax : float, positive
        Maximum dispersion for the undershoot of the hrf function.
    varDspStp : float, positive
        Step size for peak and undershoot dispersion of the hrf function.
    varPkUndRatMin : float, positive
        Minimum peak to undershoot ratio.
    varPkUndRatMax : float, positive
        Maximum peak to undershoot ratio.
    varPkUndRatStp : float, positive
        Step size for teh peak to undershoot ratio.

    Returns
    -------
    aryPrm : numpy array
        All possible hrf parameters combinations.

    """

    # Define vector for delay to peak, default: 6
    vecPkDel = np.arange(varPkDelMin, varPkDelMax+varDelStp, varDelStp)

    # Define vector for delay to undershoot, default: 16
    vecUndDel = np.arange(varUndDelMin, varUndDelMax+varDelStp, varDelStp)

    # Define vector for peak dispersion, default: 1
    vecPkDsp = np.arange(varPkDspMin, varPkDspMax+varDspStp, varDspStp)

    # Define vector for undershoot dispersion, default: 1
    vecUndDsp = np.arange(varUndDspMin, varUndDspMax+varDspStp, varDspStp)

    # Define vector for weighting of undershoot relative to peak,
    # e.g. 6 means 1:6 weighting
    vecPkUndRat = np.arange(varPkUndRatMin, varPkUndRatMax+varPkUndRatStp,
                            varPkUndRatStp)

    # Find combinations of all parameters
    # Exclude combinations where undershoot delay less than peak delay
    iterables = [vecPkDel, vecUndDel]
    vecDelCmb = list(itertools.product(*iterables))
    vecDelCmb = np.asarray(vecDelCmb)
    # Pass only the combinations where a1 < a2
    lgcDelCmb = np.less(vecDelCmb[:, 0], vecDelCmb[:, 1])
    vecDelCmb = vecDelCmb[lgcDelCmb]

    iterables = [vecDelCmb, vecPkDsp, vecUndDsp, vecPkUndRat]
    aryPrm = list(itertools.product(*iterables))
    for ind, item in enumerate(aryPrm):
        aryPrm[ind] = list(np.hstack(item))
    aryPrm = np.array(aryPrm)

    return aryPrm.astype(np.float32)


def wrp_hrf_prm_dct(vecPrm, lgcNrm=True):
    """Wrap hrf parameters into dictionary.

    Parameters
    ----------
    vecPrm : 1D numpy array
        Array with a specific combination of hrf parameters.
    lgcNrm : boolean
        Should the hrf function be passed the argument to be normalized?

    Returns
    -------
    dctPrms : dictionary
        Dictionary with hrf parameters.

    """

    # Create dictionary for hrf function
    dctPrms = {}
    dctPrms['peak_delay'] = float(vecPrm[0])
    dctPrms['under_delay'] = float(vecPrm[1])
    dctPrms['peak_disp'] = float(vecPrm[2])
    dctPrms['under_disp'] = float(vecPrm[3])
    dctPrms['p_u_ratio'] = float(vecPrm[4])

    if lgcNrm:
        dctPrms['normalize'] = True

    return dctPrms


def cvrt_hrf_prm_fn(aryPrm, fnHrf, varTr, varTmpOvsmpl, varHrfLen=32.):
    """Convert hrf parameters to hrf base function.

    Parameters
    ----------
    aryPrm : numpy array
        Array with all combinations of hrf parameters.
    fnHrf : function
        Specific hrf function that should be used
    varTr : float, positive
        Time to repeat (TR) of the (fMRI) experiment.
    varTmpOvsmpl : float, positive
        Factor by which the time courses should be temporally upsampled.
    varHrfLen : float, positive, default=32
        Length of the HRF time course in seconds.

    Returns
    -------
    aryHrfBse : 2D numpy array
        Array with hrf base functions.

    """

    # Derive array for time
    aryTme = np.linspace(0, varHrfLen, (varHrfLen // varTr) * varTmpOvsmpl)
    # Initialize array that will collect hrf base functions
    aryHrfBse = np.zeros((aryPrm.shape[0], aryTme.shape[0]), dtype=np.float32)

    # Loop over hrf parameter combinations
    for indPrm, vecPrm in enumerate(aryPrm):
        # obtain dictionary for this hrf parameter combination
        dctPrms = wrp_hrf_prm_dct(vecPrm)
        # Create hrf base function based in the parameters
        vecTmpBse = fnHrf(aryTme, **dctPrms)
        # Normalise HRF so that the sum of values is 1 (see FSL)
        # otherwise, after convolution values for predictors are very high
        vecTmpBse = np.divide(vecTmpBse, np.sum(vecTmpBse))
        # Assign normalized base function to array
        aryHrfBse[indPrm, :] = vecTmpBse

    return aryHrfBse


def cnvl_nrl_hrf(aryNrlRsp, aryHrfBse, vecFrms, vecFrmTms, varTr, varNumVol):
    """Convolution, un-vectorized, requires less RAM.

    Parameters
    ----------
    aryNrlRsp : 2D numpy array
        Neural response function(s) that won for this voxel in first fit.
    aryHrfBse : 2D numpy array
        All basic hrf functions.
    vecFrms : 1D numpy array
        Frame times, i.e. start point of every volume in seconds.
    vecFrmTms : 2D numpy array
        Supersampled frames times, i.e. start point of every volume in
        upsampled res.
    varTr : float, positive
        Time to repeat (TR) of the (fMRI) experiment.
    varNumVol : float, positive
        Number of volumes of the (fMRI) data.

    Returns
    -------
    aryConv : 2D numpy array
        Neural response function convolved with all possible hrf functions.

    """

    # Prepare an empty array for convolution ouput
    aryConv = np.zeros((aryHrfBse.shape[0], aryNrlRsp.shape[0], varNumVol),
                       dtype=np.float32)

    # Loop over neural responses
    for indNrlRsp, vecNrlRsp in enumerate(aryNrlRsp):

        # Convolving the best-fitting neural response with all hrf base fn
        for indBse, vecHrfBse in enumerate(aryHrfBse):
            # Make sure hrf basis function is float64 to avoid overflow
            vecHrfBse = vecHrfBse.astype(np.float64)
            # Perform the convolution
            col = fftconvolve(vecHrfBse, vecNrlRsp,
                              mode='full')[:vecNrlRsp.size]
            # Get function for downsampling
            f = interp1d(vecFrmTms, col)
            # Downsample to original resoltuion to match res of data
            # take the value from the centre of each volume's period (see FSL)
            aryConv[indBse, indNrlRsp, :] = f(vecFrms + varTr/2.)

    return aryConv


def cnvl_nrl_hrf_vec(aryNrlRsp, aryHrfBse, vecFrms, vecFrmTms, varTr,
                     varNumVol):
    """Convolution, vectorized but requires more RAM.

    Parameters
    ----------
    aryNrlRsp : 2D numpy array
        Neural response function(s) that won for this voxel in first fit.
    aryHrfBse : 2D numpy array
        All basic hrf functions.
    vecFrms : 1D numpy array
        Frame times, i.e. start point of every volume in seconds.
    vecFrmTms : 2D numpy array
        Supersampled frames times, i.e. start point of every volume in
        upsampled res.
    varTr : float, positive
        Time to repeat (TR) of the (fMRI) experiment.
    varNumVol : float, positive
        Number of volumes of the (fMRI) data.

    Returns
    -------
    aryConv : 2D numpy array
        Neural response function convolved with all possible hrf functions.

    """

    # Prepare an empty array for convolution ouput
    aryConv = np.zeros((aryHrfBse.shape[0], aryNrlRsp.shape[0], varNumVol),
                       dtype=np.float32)

    # Loop over neural responses
    for indNrlRsp, vecNrlRsp in enumerate(aryNrlRsp):

        # Perform the convolution
        col = fftconvolve(aryHrfBse, vecNrlRsp[None, :],
                          mode='full')[:, :vecNrlRsp.size]
        # Get function for downsampling
        f = interp1d(vecFrmTms, col)
        # Downsample to original resoltuion to match res of data
        # take the value from the centre of each volume's period (see FSL)
        aryConv[:, indNrlRsp, :] = f(vecFrms + varTr/2.)

    return aryConv.astype(np.float32)


class cls_set_config(object):
    """
    Set config parameters from dictionary into local namespace.

    Parameters
    ----------
    dicCnfg : dict
        Dictionary containing parameter names (as keys) and parameter values
        (as values). For example, `dicCnfg['varTr']` contains a float, such as
        `2.94`.
    """

    def __init__(self, dicCnfg):
        """Set config parameters from dictionary into local namespace."""
        self.__dict__.update(dicCnfg)



## Create hrf parameter combinations
#aryPrm = create_hrf_params(varPkDelMin, varPkDelMax, varUndDelMin,
#                           varUndDelMax, varDelStp, varPkDspMin, varPkDspMax,
#                           varUndDspMin, varUndDspMax, varDspStp,
#                           varPkUndRatMin, varPkUndRatMax, varPkUndRatStp)
#
## Turn one parameter combination into dictionary
#dctPrms = wrap_hrf_params_dct(aryPrm, 0)
#
## Obtain hrf function with these specific hrf parameters
#t = np.arange(30)
#testBase = spm_hrf_compat(t, **dctPrms)




# -*- coding: utf-8 -*-
"""Optimize the hrf parameters, given estimated pRF parameters."""

# Part of hrf_opt library
# Copyright (C) 2018  Marian Schneider
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
import os
import numpy as np
import multiprocessing as mp

from hrf_opt.load_config import load_config
from hrf_opt.hrf_opt_utils import cls_set_config, crt_hrf_prm, cvrt_hrf_prm_fn
from hrf_opt.find_opt_hrf import find_opt_hrf

from pyprf_feature.analysis.utils_general import (load_res_prm, joinRes,
                                                  export_nii)
from pyprf_feature.analysis.prepare import prep_func
from pyprf_feature.analysis.model_creation_utils import crt_nrl_tc
from pyprf_feature.analysis.utils_hrf import spm_hrf_compat

###### DEBUGGING ###############
#strCsvCnfg = "/media/sf_D_DRIVE/MotDepPrf/Analysis/S02/04_motDepPrf/pRF_results/Avg/S02_config_hrf_opt_flck_smooth_avg.csv"
#lgcTest = False
################################


def hrf_opt_run(strCsvCnfg, lgcTest=False):
    """
    Main function for hrf_opt.

    Parameters
    ----------
    strCsvCnfg : str
        Absolute file path of config file.
    lgcTest : Boolean
        Whether this is a test (pytest). If yes, absolute path of pyprf libary
        will be prepended to config file paths.

    """

    # %% Preparations

    # Check time
    print('---pRF analysis')
    varTme01 = time.time()

    # Load config parameters from csv file into dictionary:
    dicCnfg = load_config(strCsvCnfg, lgcTest=lgcTest)

    # Load config parameters from dictionary into namespace:
    cfg = cls_set_config(dicCnfg)

    # If suppressive surround flag is on, make sure to retrieve results from
    # that fitting
    if cfg.lgcSupsur:
        cfg.strPathOut = cfg.strPathOut + '_supsur'
        cfg.strPathFitRes = cfg.strPathFitRes + '_supsur'

    # Convert preprocessing parameters (for temporal smoothing)
    # from SI units (i.e. [s]) into units of data array (volumes):
    cfg.varSdSmthTmp = np.divide(cfg.varSdSmthTmp, cfg.varTr)

    # Check whether we need to crossvalidate
    if np.greater(cfg.varNumXval, 1):
        cfg.lgcXval = True
    elif np.equal(cfg.varNumXval, 1):
        cfg.lgcXval = False
    strErrMsg = 'Stopping program. ' + \
        'Set numXval (number of crossvalidation folds) to 1 or higher'
    assert np.greater_equal(cfg.varNumXval, 1), strErrMsg

    # %% Load empirical voxel time courses and pRF models from first fit

    # Load the empirical voxel time courses
    # Some voxels were excluded because they did not have sufficient mean
    # and/or variance - exclude their initial parameters, too
    # Get inclusion mask and nii header
    aryLgcMsk, aryLgcVar, hdrMsk, aryAff, aryFunc, tplNiiShp = prep_func(
        cfg.strPathNiiMask, cfg.lstPathNiiFunc, varAvgThr=-100)

    # Derive path to R2 results from first fit
    lstR2Path = [cfg.strPathFitRes + '_R2.nii.gz']
    # Check if fitting has been performed, i.e. whether R2 file exists
    # Throw error message if it does not exist.
    errorMsg = 'Files that should have resulted from fitting do not exist. \
                \nPlease perform pRF fitting first, calling  e.g.: \
                \npyprf_feature -config /path/to/my_config_file.csv'
    # Check if fitting has been performed, i.e. whether parameter file exists
    assert os.path.isfile(lstR2Path[0]), errorMsg
    # Load the R2 parameters from first fit
    vecR2 = load_res_prm(lstR2Path, lstFlsMsk=[cfg.strPathNiiMask])[0][0]
    # Make sure that R2 is one-dimensional
    vecR2 = np.squeeze(vecR2)
    # Apply inclusion mask to vecR2
    vecR2 = vecR2[aryLgcVar]

    # Derive path to aperture responses for winner model from first fit
    strMdlRspPath = cfg.strPathFitRes + '_FitMdlRsp.npy'
    # Check if winner model responses to aperture conditions have been saved,
    # i.e. pyprf_feature was run with -save_tc -mdl_rsp flags after fitting
    errorMsg = 'Files that should have resulted from fitting do not exist. \
                \nPlease perform pRF fitting first, then call to \
                pyprf_feature with the -save_tc and -mdl_rsp flags: \
                \npyprf_feature -config /path/to/config.csv -save_tc -mdl_rsp'
    # Check if fitting has been performed, i.e. whether response file exists
    assert os.path.isfile(strMdlRspPath), errorMsg
    # Load response to apertures for each voxel for winner model from first fit
    aryMdlRsp = np.load(strMdlRspPath)

    # Obtain a mask for voxel inclusion by thresholding with R2 value
    # Find R2 value for top varThrNumVox voxel
    varThrR2 = vecR2[np.argsort(vecR2)[::-1][cfg.varThrNumVox]]
    aryLgcR2 = np.greater(vecR2, varThrR2)
    # Apply the R2 mask to aryFunc and aryMdlRsp
    aryFunc = aryFunc[aryLgcR2, :]
    aryMdlRsp = aryMdlRsp[aryLgcR2, ...]
    # Apply the R2 mask to voxel inclusion mask
    aryLgcVar[aryLgcVar] = aryLgcR2

    # %% Prepare hrf base functions and predicted neural time courses

    # Print statement to user
    print('---Create predicted neural responses')

    # Create hrf parameter combinations
    aryPrm = crt_hrf_prm(cfg.varPkDelMin, cfg.varPkDelMax,
                         cfg.varUndDelMin, cfg.varUndDelMax,
                         cfg.varDelStp, cfg.varPkDspMin, cfg.varPkDspMax,
                         cfg.varUndDspMin, cfg.varUndDspMax, cfg.varDspStp,
                         cfg.varPkUndRatMin, cfg.varPkUndRatMax,
                         cfg.varPkUndRatStp)

    # create hrf base functions for each of the model parameter combinations
    aryHrfBse = cvrt_hrf_prm_fn(aryPrm, spm_hrf_compat, cfg.varTr,
                                cfg.varTmpOvsmpl)
    # Exclude hrf basis functions that have any NaN values
    lgcNan = np.any(np.isnan(aryHrfBse), axis=1)
    aryHrfBse = aryHrfBse[np.invert(lgcNan), :]
    aryPrm = aryPrm[np.invert(lgcNan), :]
    if np.any(lgcNan):
        print('------Nan values detected & excluded in basis hrf functions')

    # Load temporal information about when which apertures was presented in the
    # experiment
    aryTmpExpInf = np.load(cfg.strTmpExpInf)

    # Create predicted neural response for model in question
    aryNrlRsp = crt_nrl_tc(aryMdlRsp, aryTmpExpInf[:, 0], aryTmpExpInf[:, 1],
                           aryTmpExpInf[:, 2], cfg.varTr, cfg.varNumVol,
                           cfg.varTmpOvsmpl)

    # free some memory
    del(aryTmpExpInf)
    del(aryMdlRsp)

    # %% Optimize the hrf function

    # Empty list for results (parameters of best fitting pRF model):
    lstPrfRes = [None] * cfg.varPar

    # Empty list for processes:
    lstPrcs = [None] * cfg.varPar

    # Create a queue to put the results in:
    queOut = mp.Queue()

    # Print statement to user
    print('---Optimize hrf functions')
    print('------Optimization will be performed on this number of voxels: ' +
          str(aryFunc.shape[0]))
    print('------Optimization will be performed on this number of models: ' +
          str(aryPrm.shape[0]))

    # Create list with chunks of functional data for the parallel processes:
    lstFunc = np.array_split(aryFunc, cfg.varPar)
    # Create list with chunks of aryNrlRsp for the parallel processes:
    lstNrlRsp = np.array_split(aryNrlRsp, cfg.varPar)

    # We don't need the original array with the functional data anymore:
    del(aryFunc)
    del(aryNrlRsp)

    # Create processes:
    for idxPrc in range(0, cfg.varPar):
        lstPrcs[idxPrc] = mp.Process(target=find_opt_hrf,
                                     args=(idxPrc,
                                           lstFunc[idxPrc].T,
                                           lstNrlRsp[idxPrc],
                                           aryHrfBse,
                                           aryPrm,
                                           cfg.varTr,
                                           cfg.varNumVol,
                                           cfg.varTmpOvsmpl,
                                           cfg.lgcXval,
                                           cfg.varNumXval,
                                           queOut),
                                     )
        # Daemon (kills processes when exiting):
        lstPrcs[idxPrc].Daemon = True

    # Start processes:
    for idxPrc in range(0, cfg.varPar):
        lstPrcs[idxPrc].start()

    # Delete reference to list with function data (the data continues to exists
    # in child process):
    del(lstFunc)

    # Collect results from queue:
    for idxPrc in range(0, cfg.varPar):
        lstPrfRes[idxPrc] = queOut.get(True)

    # Join processes:
    for idxPrc in range(0, cfg.varPar):
        lstPrcs[idxPrc].join()

    # %% Prepare pRF finding results for export

    # Put output into correct order:
    lstPrfRes = sorted(lstPrfRes)

    # collect results from parallelization
    aryBstPrm = joinRes(lstPrfRes, cfg.varPar, 1, inFormat='2D')
    aryBstR2 = joinRes(lstPrfRes, cfg.varPar, 2, inFormat='1D')
    aryBstBts = joinRes(lstPrfRes, cfg.varPar, 3, inFormat='2D')
    aryAllRes = joinRes(lstPrfRes, cfg.varPar, 4, inFormat='2D')
    if np.greater(cfg.varNumXval, 1):
        aryBstR2Single = joinRes(lstPrfRes, cfg.varPar, 5, inFormat='2D')

    # Delete unneeded large objects:
    del(lstPrfRes)

    # %% Save results to disk:

    # List with name suffices of output images:
    lstNiiNames = ['_R2']

    # Create full path names from nii file names and output path
    lstNiiNames = [cfg.strPathOut + strNii + '.nii.gz' for strNii in
                   lstNiiNames]

    # export beta parameter as a single 4D nii file
    export_nii(np.atleast_2d(aryBstR2).T, lstNiiNames, aryLgcMsk, aryLgcVar,
               tplNiiShp, aryAff, hdrMsk, outFormat='3D')

    # List with name suffices of output images:
    lstNiiNames = ['_HrfPrm']

    # Create full path names from nii file names and output path
    lstNiiNames = [cfg.strPathOut + strNii + '.nii.gz' for strNii in
                   lstNiiNames]

    # export beta parameter as a single 4D nii file
    export_nii(aryBstPrm, lstNiiNames, aryLgcMsk, aryLgcVar, tplNiiShp,
               aryAff, hdrMsk, outFormat='4D')

    # List with name suffices of output images:
    lstNiiNames = ['_Betas']

    # Create full path names from nii file names and output path
    lstNiiNames = [cfg.strPathOut + strNii + '.nii.gz' for strNii in
                   lstNiiNames]

    # export beta parameter as a single 4D nii file
    export_nii(aryBstBts, lstNiiNames, aryLgcMsk, aryLgcVar, tplNiiShp,
               aryAff, hdrMsk, outFormat='4D')

    if np.greater(cfg.varNumXval, 1):

        # truncate extremely negative R2 values
        aryBstR2Single[np.where(np.less_equal(aryBstR2Single, -1.0))] = -1.0

        # List with name suffices of output images:
        lstNiiNames = ['_R2_single']

        # Create full path names from nii file names and output path
        lstNiiNames = [cfg.strPathOut + strNii + '.nii.gz' for strNii in
                       lstNiiNames]

        # export R2 maps as a single 4D nii file
        export_nii(aryBstR2Single, lstNiiNames, aryLgcMsk, aryLgcVar,
                   tplNiiShp, aryAff, hdrMsk, outFormat='4D')

    # %% Save global winner hrf parameter combinations

    # Save the hrf parameter combination that minimizes across all voxels
    arySumRes = np.sum(aryAllRes, axis=0)
    vecMinPrm = aryPrm[np.argmin(arySumRes, axis=0), :]
    np.save(cfg.strPathOut + '_minHrfPrm', vecMinPrm)

    # Save the average hrf parameter combination across all voxels
    vecAvgPrm = np.mean(aryBstPrm, axis=0)
    np.save(cfg.strPathOut + '_avgHrfPrm', vecAvgPrm)

    # %% Report time

    varTme02 = time.time()
    varTme03 = varTme02 - varTme01
    print('---Elapsed time: ' + str(varTme03) + ' s')
    print('---Done.')



"""
hrf_opt setup.

For development installation:
    pip install -e /path/to/hrf_opt
"""

from setuptools import setup

with open('README.rst') as f:
    long_description = f.read()

setup(name='hrf_opt',
      version='1.0.3',
      description=('Optimize hemodynamic response function parameters.'),
      url='https://github.com/MSchnei/hrf_opt',
      author='Marian Schneider',
      author_email='marian.schneider@maastrichtuniversity.nl',
      license='GNU General Public License Version 3',
      install_requires=['numpy', 'scipy', 'nibabel', 'cython==0.27.1',
                        'scikit-learn==0.19.1', 'pyprf_feature==1.1.1'],
      keywords=['pRF', 'fMRI', 'retinotopy'],
      long_description=long_description,
      packages=['hrf_opt'],
      entry_points={
          'console_scripts': [
              'hrf_opt = hrf_opt.__main__:main',
              ]},
      )



# -*- coding: utf-8 -*-
"""Load hrf_opt config file."""

# Part of hrf_opt library
# Copyright (C) 2018  Marian Schneider
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import csv
import ast

# Get path of this file:
strDir = os.path.dirname(os.path.abspath(__file__))


def load_config(strCsvCnfg, lgcTest=False, lgcPrint=True):
    """
    Load py_pRF_mapping config file.

    Parameters
    ----------
    strCsvCnfg : string
        Absolute file path of config file.
    lgcTest : Boolean
        Whether this is a test (pytest). If yes, absolute path of this function
        will be prepended to config file paths.
    lgcPrint : Boolean
        Print config parameters?

    Returns
    -------
    dicCnfg : dict
        Dictionary containing parameter names (as keys) and parameter values
        (as values). For example, `dicCnfg['varTr']` contains a float, such as
        `2.94`.

    """

    # Dictionary with config information:
    dicCnfg = {}

    # Open file with parameter configuration:
    # fleConfig = open(strCsvCnfg, 'r')
    with open(strCsvCnfg, 'r') as fleConfig:

        # Read file  with ROI information:
        csvIn = csv.reader(fleConfig,
                           delimiter='\n',
                           skipinitialspace=True)

        # Loop through csv object to fill list with csv data:
        for lstTmp in csvIn:

            # Skip comments (i.e. lines starting with '#') and empty lines.
            # Note: Indexing the list (i.e. lstTmp[0][0]) does not work for
            # empty lines. However, if the first condition is no fullfilled
            # (i.e. line is empty and 'if lstTmp' evaluates to false), the
            # second logical test (after the 'and') is not actually carried
            # out.
            if lstTmp and not (lstTmp[0][0] == '#'):

                # Name of current parameter (e.g. 'varTr'):
                strParamKey = lstTmp[0].split(' = ')[0]
                # print(strParamKey)

                # Current parameter value (e.g. '2.94'):
                strParamVlu = lstTmp[0].split(' = ')[1]
                # print(strParamVlu)

                # Put paramter name (key) and value (item) into dictionary:
                dicCnfg[strParamKey] = strParamVlu

    # Minimum delay for the peak of the hrf function:
    dicCnfg['varPkDelMin'] = float(dicCnfg['varPkDelMin'])
    if lgcPrint:
        print('---Minimum delay for the peak of the hrf function: '
              + str(dicCnfg['varPkDelMin']))

    # Maximum delay for the peak of the hrf function:
    dicCnfg['varPkDelMax'] = float(dicCnfg['varPkDelMax'])
    if lgcPrint:
        print('---Maximum delay for the peak of the hrf function: '
              + str(dicCnfg['varPkDelMax']))

    # Minimum delay for the undershoot of the hrf function:
    dicCnfg['varUndDelMin'] = float(dicCnfg['varUndDelMin'])
    if lgcPrint:
        print('---Minimum delay for the undershoot of the hrf function: '
              + str(dicCnfg['varUndDelMin']))

    # Maximum delay for the undershoot of the hrf function:
    dicCnfg['varUndDelMax'] = float(dicCnfg['varUndDelMax'])
    if lgcPrint:
        print('---Maximum delay for the undershoot of the hrf function: '
              + str(dicCnfg['varUndDelMax']))

    # Step size for peak and undershoot delay of the hrf function:
    dicCnfg['varDelStp'] = float(dicCnfg['varDelStp'])
    if lgcPrint:
        print('---Step size for peak & undershoot delay of the hrf function: '
              + str(dicCnfg['varDelStp']))

    # Minimum dispersion for the peak of the hrf function:
    dicCnfg['varPkDspMin'] = float(dicCnfg['varPkDspMin'])
    if lgcPrint:
        print('---Minimum dispersion for the peak of the hrf function: '
              + str(dicCnfg['varPkDspMin']))

    # Maximum dispersion for the peak of the hrf function:
    dicCnfg['varPkDspMax'] = float(dicCnfg['varPkDspMax'])
    if lgcPrint:
        print('---Maximum dispersion for the peak of the hrf function: '
              + str(dicCnfg['varPkDspMax']))

    # Minimum dispersion for the undershoot of the hrf function:
    dicCnfg['varUndDspMin'] = float(dicCnfg['varUndDspMin'])
    if lgcPrint:
        print('---Minimum dispersion for the undershoot of the hrf function: '
              + str(dicCnfg['varUndDspMin']))

    # Maximum dispersion for the undershoot of the hrf function:
    dicCnfg['varUndDspMax'] = float(dicCnfg['varUndDspMax'])
    if lgcPrint:
        print('---Maximum dispersion for the undershoot of the hrf function: '
              + str(dicCnfg['varUndDspMax']))

    # Step size for peak and undershoot dispersion of the hrf function:
    dicCnfg['varDspStp'] = float(dicCnfg['varDspStp'])
    if lgcPrint:
        print('---Step size for dispersion of the hrf function: '
              + str(dicCnfg['varDspStp']))

    # Minimum peak to undershoot ratio:
    dicCnfg['varPkUndRatMin'] = float(dicCnfg['varPkUndRatMin'])
    if lgcPrint:
        print('---Minimum peak to undershoot ratio: '
              + str(dicCnfg['varPkUndRatMin']))

    # Maximum peak to undershoot ratio:
    dicCnfg['varPkUndRatMax'] = float(dicCnfg['varPkUndRatMax'])
    if lgcPrint:
        print('---Maximum peak to undershoot ratio: '
              + str(dicCnfg['varPkUndRatMax']))

    # Step size for the peak to undershoot ratio
    dicCnfg['varPkUndRatStp'] = float(dicCnfg['varPkUndRatStp'])
    if lgcPrint:
        print('---Step size for the peak to undershoot ratio: '
              + str(dicCnfg['varPkUndRatStp']))

    # Factor by which time courses and HRF will be upsampled for the
    # convolutions
    dicCnfg['varTmpOvsmpl'] = ast.literal_eval(dicCnfg['varTmpOvsmpl'])
    if lgcPrint:
        print('---Factor by which time courses and HRF will be upsampled: '
              + str(dicCnfg['varTmpOvsmpl']))

    # Extent of temporal smoothing for fMRI data and pRF time course models
    # [standard deviation of the Gaussian kernel, in seconds]:
    # same temporal smoothing will be applied to pRF model time courses
    dicCnfg['varSdSmthTmp'] = float(dicCnfg['varSdSmthTmp'])
    if lgcPrint:
        print('---Extent of temporal smoothing (Gaussian SD in [s]): '
              + str(dicCnfg['varSdSmthTmp']))

    # Volume TR of input data [s]:
    dicCnfg['varTr'] = float(dicCnfg['varTr'])
    if lgcPrint:
        print('---Volume TR of input data [s]: ' + str(dicCnfg['varTr']))

    # Number of fMRI volumes and png files to load:
    dicCnfg['varNumVol'] = int(dicCnfg['varNumVol'])
    if lgcPrint:
        print('---Total number of fMRI volumes and png files: '
              + str(dicCnfg['varNumVol']))

    # Path(s) of functional data:
    dicCnfg['lstPathNiiFunc'] = ast.literal_eval(dicCnfg['lstPathNiiFunc'])
    if lgcPrint:
        print('---Path(s) of functional data:')
        for strTmp in dicCnfg['lstPathNiiFunc']:
            print('   ' + str(strTmp))

    # Path of mask (to restrict pRF model finding):
    dicCnfg['strPathNiiMask'] = ast.literal_eval(dicCnfg['strPathNiiMask'])
    if lgcPrint:
        print('---Path of mask (to restrict pRF model finding):')
        print('   ' + str(dicCnfg['strPathNiiMask']))

    # Path to base of npy files with model parameters and responses:
    dicCnfg['strPathMdlRsp'] = ast.literal_eval(dicCnfg['strPathMdlRsp'])
    if lgcPrint:
        print('---Path to base of files with model parameters and responses: ')
        print('   ' + str(dicCnfg['strPathMdlRsp']))

    # Path to npy file with temporal info about aperture conditions
    dicCnfg['strTmpExpInf'] = ast.literal_eval(dicCnfg['strTmpExpInf'])
    if lgcPrint:
        print('---Path to npy file with temporal condition info: ')
        print('   ' + str(dicCnfg['strTmpExpInf']))

    # Were the model time courses created with suppressive surround?
    dicCnfg['lgcSupsur'] = (dicCnfg['lgcSupsur'] == 'True')
    if lgcPrint:
        print('---Model time courses were created with suppressive surround: '
              + str(dicCnfg['lgcSupsur']))

    # Path to base of nifti files with previous pRF fitting results:
    dicCnfg['strPathFitRes'] = ast.literal_eval(dicCnfg['strPathFitRes'])
    if lgcPrint:
        print('---Path to base of files with previous pRF fitting results: ')
        print('   ' + str(dicCnfg['strPathFitRes']))

    # Top number of voxels that will be considered for hrf pool:
    dicCnfg['varThrNumVox'] = float(dicCnfg['varThrNumVox'])
    if lgcPrint:
        print('---Top number of voxels that will be considered for hrf pool: '
              + str(dicCnfg['varThrNumVox']))

    # Number of processes to run in parallel:
    dicCnfg['varPar'] = int(dicCnfg['varPar'])
    if lgcPrint:
        print('---Number of processes to run in parallel: '
              + str(dicCnfg['varPar']))

    # should model fitting be based on k-fold cross-validation?
    # if not, set to 1
    dicCnfg['varNumXval'] = int(dicCnfg['varNumXval'])
    if lgcPrint:
        print('---Model fitting will have this number of folds for xval: '
              + str(dicCnfg['varNumXval']))

    # Output basename:
    dicCnfg['strPathOut'] = ast.literal_eval(dicCnfg['strPathOut'])
    if lgcPrint:
        print('---Output basename:')
        print('   ' + str(dicCnfg['strPathOut']))

    # Is this a test?
    if lgcTest:

        # Prepend absolute path of this file to config file paths:
        dicCnfg['strPathFuncIn'] = (strDir + dicCnfg['strPathFuncIn'])
        dicCnfg['strPathOut'] = (strDir + dicCnfg['strPathOut'])
        dicCnfg['strPathMdlRsp'] = (strDir + dicCnfg['strPathMdlRsp'])
        dicCnfg['strTmpExpInf'] = (strDir + dicCnfg['strTmpExpInf'])
        dicCnfg['strPathR2'] = (strDir + dicCnfg['strPathR2'])

        # Loop through functional runs:
        varNumRun = len(dicCnfg['lstPathNiiFunc'])
        for idxRun in range(varNumRun):
            dicCnfg['lstPathNiiFunc'][idxRun] = (
                strDir
                + dicCnfg['lstPathNiiFunc'][idxRun]
                )

    return dicCnfg
